@@ -18,67 +18,60 @@
 # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 # THE SOFTWARE.
 
+from collections import OrderedDict
 import glob
 import imp
+from itertools import product
 import os
+try:
+    from pathlib import Path
+except ImportError:
+    from pathlib2 import Path
+import subprocess
 
 import six
-from ansible import constants
-from ansible.errors import AnsibleError
-
-try:
-    # Try to import the Ansible 2 module first, it's the future-proof one
-    from ansible.parsing.splitter import split_args
-
-except ImportError:
-    # Fallback on the Ansible 1.9 module
-    from ansible.module_utils.splitter import split_args
-
 import yaml
 from yaml.composer import Composer
-from yaml.constructor import Constructor
+import ruamel.yaml
 
-try:
-    from ansible.utils import parse_yaml_from_file
-    from ansible.utils import path_dwim
-    from ansible.utils.template import template as ansible_template
-    from ansible.utils import module_finder
-    module_loader = module_finder
-    ANSIBLE_VERSION = 1
-except ImportError:
-    from ansible.parsing.dataloader import DataLoader
-    from ansible.template import Templar
-    from ansible.parsing.mod_args import ModuleArgsParser
-    from ansible.parsing.yaml.constructor import AnsibleConstructor
-    from ansible.parsing.yaml.loader import AnsibleLoader
-    from ansible.errors import AnsibleParserError
-    ANSIBLE_VERSION = 2
+from ansible import constants
+from ansible.errors import AnsibleError
+from ansible.errors import AnsibleParserError
+from ansible.parsing.dataloader import DataLoader
+from ansible.parsing.mod_args import ModuleArgsParser
+from ansible.parsing.splitter import split_args
+from ansible.parsing.yaml.constructor import AnsibleConstructor
+from ansible.parsing.yaml.loader import AnsibleLoader
+from ansible.parsing.yaml.objects import AnsibleSequence
+from ansible.plugins.loader import module_loader
+from ansible.template import Templar
 
-    # ansible-lint doesn't need/want to know about encrypted secrets, but it needs
-    # Ansible 2.3+ allows encrypted secrets within yaml files, so we pass a string
-    # as the password to enable such yaml files to be opened and parsed successfully.
-    DEFAULT_VAULT_PASSWORD = 'x'
 
-    def parse_yaml_from_file(filepath):
-        dl = DataLoader()
-        if hasattr(dl, 'set_vault_password'):
-            dl.set_vault_password(DEFAULT_VAULT_PASSWORD)
-        return dl.load_from_file(filepath)
+# ansible-lint doesn't need/want to know about encrypted secrets, so we pass a
+# string as the password to enable such yaml files to be opened and parsed
+# successfully.
+DEFAULT_VAULT_PASSWORD = 'x'
 
-    def path_dwim(basedir, given):
-        dl = DataLoader()
-        dl.set_basedir(basedir)
-        return dl.path_dwim(given)
 
-    def ansible_template(basedir, varname, templatevars, **kwargs):
-        dl = DataLoader()
-        dl.set_basedir(basedir)
-        templar = Templar(dl, variables=templatevars)
-        return templar.template(varname, **kwargs)
-    try:
-        from ansible.plugins import module_loader
-    except ImportError:
-        from ansible.plugins.loader import module_loader
+def parse_yaml_from_file(filepath):
+    dl = DataLoader()
+    if hasattr(dl, 'set_vault_password'):
+        dl.set_vault_password(DEFAULT_VAULT_PASSWORD)
+    return dl.load_from_file(filepath)
+
+
+def path_dwim(basedir, given):
+    dl = DataLoader()
+    dl.set_basedir(basedir)
+    return dl.path_dwim(given)
+
+
+def ansible_template(basedir, varname, templatevars, **kwargs):
+    dl = DataLoader()
+    dl.set_basedir(basedir)
+    templar = Templar(dl, variables=templatevars)
+    return templar.template(varname, **kwargs)
+
 
 LINE_NUMBER_KEY = '__line__'
 FILENAME_KEY = '__file__'
@@ -245,12 +238,14 @@ def _taskshandlers_children(basedir, k, v, parent_type):
         elif 'import_tasks' in th:
             append_children(th['import_tasks'], basedir, k, parent_type, results)
         elif 'import_role' in th:
-            results.extend(_roles_children(basedir, k, [th['import_role'].get('name')], parent_type,
-                                           main=th['import_role'].get('tasks_from', 'main')))
+            th = normalize_task_v2(th)
+            results.extend(_roles_children(basedir, k, [th['action'].get('name')], parent_type,
+                                           main=th['action'].get('tasks_from', 'main')))
         elif 'include_role' in th:
-            results.extend(_roles_children(basedir, k, [th['include_role'].get('name')],
+            th = normalize_task_v2(th)
+            results.extend(_roles_children(basedir, k, [th['action'].get('name')],
                                            parent_type,
-                                           main=th['include_role'].get('tasks_from', 'main')))
+                                           main=th['action'].get('tasks_from', 'main')))
         elif 'block' in th:
             results.extend(_taskshandlers_children(basedir, k, th['block'], parent_type))
             if 'rescue' in th:
@@ -308,7 +303,6 @@ def _rolepath(basedir, role):
             basedir, os.path.join('..', '..', '..', 'roles', role)
         ),
         path_dwim(basedir, os.path.join('..', '..', role)),
-        path_dwim(basedir, ''),
     ]
 
     if constants.DEFAULT_ROLES_PATH:
@@ -319,6 +313,8 @@ def _rolepath(basedir, role):
             loc = os.path.expanduser(loc)
             possible_paths.append(path_dwim(loc, role))
 
+    possible_paths.append(path_dwim(basedir, ''))
+
     for path_option in possible_paths:
         if os.path.isdir(path_option):
             role_path = path_option
@@ -338,11 +334,14 @@ def _look_for_role_files(basedir, role, main='main'):
     results = []
 
     for th in ['tasks', 'handlers', 'meta']:
-        for ext in ('.yml', '.yaml'):
-            thpath = os.path.join(role_path, th, main + ext)
-            if os.path.exists(thpath):
-                results.append({'path': thpath, 'type': th})
-                break
+        current_path = os.path.join(role_path, th)
+        for dir, subdirs, files in os.walk(current_path):
+            for file in files:
+                file_ignorecase = file.lower()
+                if file_ignorecase.endswith(('.yml', '.yaml')):
+                    thpath = os.path.join(dir, file)
+                    results.append({'path': thpath, 'type': th})
+
     return results
 
 
@@ -350,7 +349,7 @@ def rolename(filepath):
     idx = filepath.find('roles/')
     if idx < 0:
         return ''
-    role = filepath[idx+6:]
+    role = filepath[idx + 6:]
     role = role[:role.find('/')]
     return role
 
@@ -361,8 +360,7 @@ def _kv_to_dict(v):
 
 
 def normalize_task_v2(task):
-    '''Ensures tasks have an action key and strings are converted to python objects'''
-
+    """Ensure tasks have an action key and strings are converted to python objects."""
     result = dict()
     mod_arg_parser = ModuleArgsParser(task)
     try:
@@ -402,6 +400,11 @@ def normalize_task_v2(task):
         del(arguments['_raw_params'])
     else:
         result['action']['__ansible_arguments__'] = list()
+
+    if 'argv' in arguments and not result['action']['__ansible_arguments__']:
+        result['action']['__ansible_arguments__'] = arguments['argv']
+        del(arguments['argv'])
+
     result['action'].update(arguments)
     return result
 
@@ -458,10 +461,7 @@ def normalize_task(task, filename):
     ansible_action_type = task.get('__ansible_action_type__', 'task')
     if '__ansible_action_type__' in task:
         del(task['__ansible_action_type__'])
-    if ANSIBLE_VERSION < 2:
-        task = normalize_task_v1(task)
-    else:
-        task = normalize_task_v2(task)
+    task = normalize_task_v2(task)
     task[FILENAME_KEY] = filename
     task['__ansible_action_type__'] = ansible_action_type
     return task
@@ -534,11 +534,10 @@ def get_normalized_tasks(yaml, file):
 
 
 def parse_yaml_linenumbers(data, filename):
-    """Parses yaml as ansible.utils.parse_yaml but with linenumbers.
+    """Parse yaml as ansible.utils.parse_yaml but with linenumbers.
 
     The line numbers are stored in each node's LINE_NUMBER_KEY key.
     """
-
     def compose_node(parent, index):
         # the line number where the previous token has ended (plus empty lines)
         line = loader.line
@@ -547,10 +546,7 @@ def parse_yaml_linenumbers(data, filename):
         return node
 
     def construct_mapping(node, deep=False):
-        if ANSIBLE_VERSION < 2:
-            mapping = Constructor.construct_mapping(loader, node, deep=deep)
-        else:
-            mapping = AnsibleConstructor.construct_mapping(loader, node, deep=deep)
+        mapping = AnsibleConstructor.construct_mapping(loader, node, deep=deep)
         if hasattr(node, '__line__'):
             mapping[LINE_NUMBER_KEY] = node.__line__
         else:
@@ -559,17 +555,276 @@ def parse_yaml_linenumbers(data, filename):
         return mapping
 
     try:
-        if ANSIBLE_VERSION < 2:
-            loader = yaml.Loader(data)
-        else:
-            import inspect
-            kwargs = {}
-            if 'vault_password' in inspect.getargspec(AnsibleLoader.__init__).args:
-                kwargs['vault_password'] = DEFAULT_VAULT_PASSWORD
-            loader = AnsibleLoader(data, **kwargs)
+        import inspect
+        kwargs = {}
+        if 'vault_password' in inspect.getargspec(AnsibleLoader.__init__).args:
+            kwargs['vault_password'] = DEFAULT_VAULT_PASSWORD
+        loader = AnsibleLoader(data, **kwargs)
         loader.compose_node = compose_node
         loader.construct_mapping = construct_mapping
         data = loader.get_single_data()
     except (yaml.parser.ParserError, yaml.scanner.ScannerError) as e:
         raise SystemExit("Failed to parse YAML in %s: %s" % (filename, str(e)))
     return data
+
+
+def get_first_cmd_arg(task):
+    try:
+        if 'cmd' in task['action']:
+            first_cmd_arg = task['action']['cmd'].split()[0]
+        else:
+            first_cmd_arg = task['action']['__ansible_arguments__'][0]
+    except IndexError:
+        return None
+    return first_cmd_arg
+
+
+def append_skipped_rules(pyyaml_data, file_text, file_type):
+    """Append 'skipped_rules' to individual tasks or single metadata block.
+
+    For a file, uses 2nd parser (ruamel.yaml) to pull comments out of
+    yaml subsets, check for '# noqa' skipped rules, and append any skips to the
+    original parser (pyyaml) data relied on by remainder of ansible-lint.
+
+    :param pyyaml_data: file text parsed via ansible and pyyaml.
+    :param file_text: raw file text.
+    :param file_type: type of file: tasks, handlers or meta.
+    :returns: original pyyaml_data altered with a 'skipped_rules' list added
+    to individual tasks, or added to the single metadata block.
+    """
+    try:
+        yaml_skip = _append_skipped_rules(pyyaml_data, file_text, file_type)
+    except RuntimeError as exc:
+        # Notify user of skip error, do not stop, do not change exit code
+        print('Error trying to append skipped rules: {!r}'.format(exc))
+        return pyyaml_data
+    return yaml_skip
+
+
+def _append_skipped_rules(pyyaml_data, file_text, file_type):
+    # parse file text using 2nd parser library
+    yaml = ruamel.yaml.YAML()
+    ruamel_data = yaml.load(file_text)
+
+    if file_type == 'meta':
+        pyyaml_data[0]['skipped_rules'] = _get_rule_skips_from_yaml(ruamel_data)
+        return pyyaml_data
+
+    # create list of blocks of tasks or nested tasks
+    if file_type in ('tasks', 'handlers'):
+        ruamel_task_blocks = ruamel_data
+        pyyaml_task_blocks = pyyaml_data
+    elif file_type in ('playbook', 'pre_tasks', 'post_tasks'):
+        try:
+            pyyaml_task_blocks = _get_task_blocks_from_playbook(pyyaml_data)
+            ruamel_task_blocks = _get_task_blocks_from_playbook(ruamel_data)
+        except (AttributeError, TypeError):
+            # TODO(awcrosby): running ansible-lint on any .yml file will
+            # assume it is a playbook, check needs to be added higher in the
+            # call stack, and can remove this except
+            return pyyaml_data
+    else:
+        raise RuntimeError('Unexpected file type: {}'.format(file_type))
+
+    # get tasks from blocks of tasks
+    pyyaml_tasks = _get_tasks_from_blocks(pyyaml_task_blocks)
+    ruamel_tasks = _get_tasks_from_blocks(ruamel_task_blocks)
+
+    # append skipped_rules for each task
+    for ruamel_task, pyyaml_task in zip(ruamel_tasks, pyyaml_tasks):
+        if pyyaml_task.get('name') != ruamel_task.get('name'):
+            raise RuntimeError('Error in matching skip comment to a task')
+        pyyaml_task['skipped_rules'] = _get_rule_skips_from_yaml(ruamel_task)
+
+    return pyyaml_data
+
+
+def _get_task_blocks_from_playbook(playbook):
+    """Return parts of playbook that contains tasks, and nested tasks.
+
+    :param playbook: playbook yaml from yaml parser.
+    :returns: list of task dictionaries.
+    """
+    PLAYBOOK_TASK_KEYWORDS = [
+        'tasks',
+        'pre_tasks',
+        'post_tasks',
+        'handlers',
+    ]
+
+    task_blocks = []
+    for play, key in product(playbook, PLAYBOOK_TASK_KEYWORDS):
+        task_blocks.extend(play.get(key, []))
+    return task_blocks
+
+
+def _get_tasks_from_blocks(task_blocks):
+    """Get list of tasks from list made of tasks and nested tasks."""
+    NESTED_TASK_KEYS = [
+        'block',
+        'always',
+        'rescue',
+    ]
+
+    def get_nested_tasks(task):
+        return (
+            subtask
+            for k in NESTED_TASK_KEYS if k in task
+            for subtask in task[k]
+        )
+
+    for task in task_blocks:
+        for sub_task in get_nested_tasks(task):
+            yield sub_task
+        yield task
+
+
+def _get_rule_skips_from_yaml(yaml_input):
+    """Travese yaml for comments with rule skips and return list of rules."""
+    def traverse_yaml(obj):
+        yaml_comment_obj_strs.append(str(obj.ca.items))
+        if isinstance(obj, dict):
+            for key, val in obj.items():
+                if isinstance(val, (dict, list)):
+                    traverse_yaml(val)
+        elif isinstance(obj, list):
+            for e in obj:
+                if isinstance(e, (dict, list)):
+                    traverse_yaml(e)
+        else:
+            return
+
+    yaml_comment_obj_strs = []
+    traverse_yaml(yaml_input)
+
+    rule_id_list = []
+    for comment_obj_str in yaml_comment_obj_strs:
+        for line in comment_obj_str.split('\\n'):
+            rule_id_list.extend(get_rule_skips_from_line(line))
+
+    return rule_id_list
+
+
+def get_rule_skips_from_line(line):
+    rule_id_list = []
+    if '# noqa' in line:
+        noqa_text = line.split('# noqa')[1]
+        rule_id_list = noqa_text.split()
+    return rule_id_list
+
+
+def normpath(path):
+    """
+    Normalize a path in order to provide a more consistent output.
+
+    Currently it generates a relative path but in the future we may want to
+    make this user configurable.
+    """
+    # convertion to string in order to allow receiving non string objects as
+    # arguments which would have failed under python2
+    return os.path.relpath(str(path))
+
+
+def is_playbook(filename):
+    """
+    Check if the file is a playbook.
+
+    Given a filename, it should return true if it looks like a playbook. The
+    function is not supposed to raise exceptions.
+    """
+    # we assume is a playbook if we loaded a sequence of dictionaries where
+    # at least one of these keys is present:
+    playbooks_keys = {
+        "gather_facts",
+        "hosts",
+        "import_playbook",
+        "post_tasks",
+        "pre_tasks",
+        "roles"
+        "tasks",
+    }
+
+    # makes it work with Path objects by converting them to strings
+    if not isinstance(filename, six.string_types):
+        filename = str(filename)
+
+    try:
+        f = parse_yaml_from_file(filename)
+    except Exception as e:
+        print(
+            "Warning: Failed to load %s with %s, assuming is not a playbook."
+            % (filename, e))
+    else:
+        if (
+            isinstance(f, AnsibleSequence) and
+            playbooks_keys.intersection(next(iter(f), {}).keys())
+        ):
+            return True
+    return False
+
+
+def get_playbooks_and_roles(options=None):
+    """Find roles and playbooks."""
+    if options is None:
+        options = {}
+
+    # git is preferred as it also considers .gitignore
+    files = OrderedDict.fromkeys(sorted(subprocess.check_output(
+        ["git", "ls-files", "*.yaml", "*.yml"],
+        universal_newlines=True).split()))
+
+    playbooks = []
+    role_dirs = []
+    role_internals = {
+        'defaults',
+        'files',
+        'handlers',
+        'meta',
+        'tasks',
+        'templates',
+        'vars',
+    }
+
+    # detect role in repository root:
+    if 'tasks/main.yml' in files or 'tasks/main.yaml' in files:
+        role_dirs.append('.')
+
+    for p in map(Path, files):
+
+        if any(str(p).startswith(file_path) for file_path in options.exclude_paths):
+            continue
+        elif (next((i for i in p.parts if i.endswith('playbooks')), None) or
+                'playbook' in p.parts[-1]):
+            playbooks.append(normpath(p))
+            continue
+
+        # ignore if any folder ends with _vars
+        if next((i for i in p.parts if i.endswith('_vars')), None):
+            continue
+        elif 'roles' in p.parts or '.' in role_dirs:
+            if 'tasks' in p.parts and p.parts[-1] in ['main.yaml', 'main.yml']:
+                role_dirs.append(str(p.parents[1]))
+            elif role_internals.intersection(p.parts):
+                continue
+            elif 'tests' in p.parts:
+                playbooks.append(normpath(p))
+        if 'molecule' in p.parts:
+            if p.parts[-1] != 'molecule.yml':
+                playbooks.append(normpath(p))
+            continue
+        # hidden files are clearly not playbooks, likely config files.
+        if p.parts[-1].startswith('.'):
+            continue
+
+        if is_playbook(p):
+            playbooks.append(normpath(p))
+            continue
+
+        if options.verbosity:
+            print('Unknown file type: %s' % normpath(p))
+
+    if options.verbosity:
+        print('Found roles: ' + ' '.join(role_dirs))
+        print('Found playbooks: ' + ' '.join(playbooks))
+
+    return role_dirs + playbooks
